"""
Entrypoint for brewblox_mdns
"""

import asyncio
import logging
import re
import sys
from glob import glob

import click
from brewblox_service import brewblox_logger, service

from brewblox_mdns import dns_discovery

LOGGER = brewblox_logger(__name__)


def print_usb():
    lines = '\n'.join([f for f in glob('/dev/serial/by-id/*')])
    for obj in re.finditer(r'particle_(?P<model>p1|photon)_(?P<serial>[a-z0-9]+)-',
                           lines,
                           re.IGNORECASE | re.MULTILINE):
        print('usb', obj.group('serial'), obj.group('model'))


async def print_wifi():
    async for res in dns_discovery.discover_all(None,
                                                dns_discovery.BREWBLOX_DNS_TYPE,
                                                dns_discovery.DEFAULT_TIMEOUT_S):
        host, port, serial = res
        print('wifi', serial, host, port)


@click.command()
@click.option('--cli', is_flag=True, help='Triggers CLI mode')
@click.option('--discovery',
              type=click.Choice(['all', 'usb', 'wifi']),
              default='all',
              help='Discovery setting. Use "all" to check both Wifi and USB')
def cli(cli, discovery):
    if discovery in ['all', 'usb']:
        print_usb()

    if discovery in ['all', 'wifi']:
        asyncio.run(print_wifi())


def main(args=sys.argv):
    if '--cli' in args:
        cli()
        return

    app = service.create_app(default_name='mdns')
    logging.captureWarnings(True)

    dns_discovery.setup(app)

    service.furnish(app)
    service.run(app)


if __name__ == '__main__':
    main()



"""
Tests brewblox_mdns.__main__
"""

from unittest.mock import call

from click.testing import CliRunner

from brewblox_mdns import __main__ as main

TESTED = main.__name__


def test_main(mocker, app):
    mocker.patch(TESTED + '.service.run')
    mocker.patch(TESTED + '.service.create_app').return_value = app
    main.main([])


def test_cli_start(mocker):
    m = mocker.patch(TESTED + '.cli')
    main.main(['--cli'])
    assert m.call_count == 1


def test_print_usb(mocker):
    entry = 'usb-Particle_P1_4f0052000551353432383931-if00'
    m = mocker.patch(TESTED + '.glob', return_value=[entry]*2)
    m_print = mocker.patch(TESTED + '.print')

    runner = CliRunner()
    assert not runner.invoke(main.cli, ['--cli', '--discovery=usb']).exception

    assert m.call_count == 1
    assert m_print.call_args_list == [
        call('usb', '4f0052000551353432383931', 'P1'),
        call('usb', '4f0052000551353432383931', 'P1'),
    ]


def test_print_wifi(mocker):
    async def discovery_generator(id, type, timeout):
        for i in range(5):
            yield f'addr-{i}', '8332', 'serial-1'

    mocker.patch(TESTED + '.dns_discovery.discover_all', discovery_generator)
    m_print = mocker.patch(TESTED + '.print')

    runner = CliRunner()
    assert not runner.invoke(main.cli, ['--cli', '--discovery=wifi']).exception

    assert m_print.call_args_list == [
        call('wifi', 'serial-1', 'addr-0', '8332'),
        call('wifi', 'serial-1', 'addr-1', '8332'),
        call('wifi', 'serial-1', 'addr-2', '8332'),
        call('wifi', 'serial-1', 'addr-3', '8332'),
        call('wifi', 'serial-1', 'addr-4', '8332'),
    ]



"""
Tests brewblox_mdns.dns_discovery
"""

import asyncio
from socket import inet_aton

import pytest
from aiohttp.client_exceptions import ContentTypeError
from aiozeroconf import ServiceInfo, ServiceStateChange

from brewblox_mdns import dns_discovery

TESTED = dns_discovery.__name__


async def response(request, status=200):
    retv = await request
    if retv.status != status:
        print(retv)
        assert retv == status
    try:
        return await retv.json()
    except ContentTypeError:
        return await retv.text()


class ServiceBrowserMock():

    def __init__(self, conf, service_type, handlers):
        print(conf, service_type, handlers)
        self.conf = conf
        self.service_type = service_type
        self.handlers = handlers

        for name in ['id0', 'id1', 'id2']:
            self.handlers[0](conf, service_type, name, ServiceStateChange.Added)
            self.handlers[0](conf, service_type, name, ServiceStateChange.Removed)


@pytest.fixture
async def app(app):
    dns_discovery.setup(app)
    return app


@pytest.fixture
def conf_mock(mocker):

    async def get_service_info(service_type, name):
        print(service_type, name)
        dns_type = dns_discovery.BREWBLOX_DNS_TYPE
        if name == 'id0':
            return ServiceInfo(
                service_type,
                f'{name}.{dns_type}',
                address=inet_aton('0.0.0.0')
            )
        if name == 'id1':
            return ServiceInfo(
                service_type,
                f'{name}.{dns_type}',
                server=f'{name}.local.',
                address=inet_aton('1.2.3.4'),
                port=1234
            )
        if name == 'id2':
            return ServiceInfo(
                service_type,
                f'{name}.{dns_type}',
                server=f'{name}.local.',
                address=inet_aton('4.3.2.1'),
                port=4321
            )

    async def close():
        pass

    m = mocker.patch(TESTED + '.Zeroconf')
    m.return_value.get_service_info = get_service_info
    m.return_value.close = close
    return m


@pytest.fixture
def browser_mock(mocker):
    return mocker.patch(TESTED + '.ServiceBrowser', ServiceBrowserMock)


async def test_discover_one(app, client, loop, browser_mock, conf_mock):
    dns_type = dns_discovery.BREWBLOX_DNS_TYPE
    assert await dns_discovery.discover_one(None, dns_type) == ('1.2.3.4', 1234, 'id1')
    assert await dns_discovery.discover_one('id2', dns_type) == ('4.3.2.1', 4321, 'id2')

    assert await dns_discovery.discover_one(None, dns_type, 1) == ('1.2.3.4', 1234, 'id1')
    assert await dns_discovery.discover_one('id2', dns_type, 1) == ('4.3.2.1', 4321, 'id2')

    with pytest.raises(asyncio.TimeoutError):
        await dns_discovery.discover_one(loop, 'leprechauns', 0.1)


async def test_discover_all(app, client, loop, browser_mock, conf_mock):
    dns_type = dns_discovery.BREWBLOX_DNS_TYPE
    retv = []
    async for res in dns_discovery.discover_all(None, dns_type, 0.01):
        retv.append(res)
    assert len(retv) == 2


async def test_post_discover(app, client, loop, browser_mock, conf_mock):
    resp1 = {
        'host': '1.2.3.4',
        'port': 1234,
        'id': 'id1',
    }
    resp2 = {
        'host': '4.3.2.1',
        'port': 4321,
        'id': 'id2',
    }

    assert await response(client.post('/discover', json={})) == resp1
    assert await response(client.post('/discover', json={'id': 'ID2'})) == resp2
    assert await response(client.post('/discover_all', json={'timeout': 0.01})) == [resp1, resp2]
